package frc.robot.subsystems;

import edu.wpi.first.epilogue.Logged;
import edu.wpi.first.epilogue.Logged.Importance;
import edu.wpi.first.math.estimator.MecanumDrivePoseEstimator;
import edu.wpi.first.math.geometry.Pose2d;
import edu.wpi.first.math.geometry.Rotation2d;
import edu.wpi.first.math.kinematics.ChassisSpeeds;
import edu.wpi.first.math.kinematics.MecanumDriveWheelPositions;
import edu.wpi.first.math.kinematics.MecanumDriveWheelSpeeds;
import edu.wpi.first.wpilibj.ADXRS450_Gyro;
import edu.wpi.first.wpilibj.Encoder;
import edu.wpi.first.wpilibj.drive.MecanumDrive;
import edu.wpi.first.wpilibj.motorcontrol.VictorSP;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.PrintCommand;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import edu.wpi.first.wpilibj2.command.button.CommandXboxController;
import frc.robot.constants.MecanumDriveConstants;
import frc.robot.interfaces.IDrivetrain;

public class MecanumDrivetrain extends SubsystemBase implements IDrivetrain {
    private VictorSP frontLeft;
    private VictorSP rearLeft;
    private VictorSP frontRight;
    private VictorSP rearRight;

    private MecanumDrive drive;

    private Encoder frontLeftEncoder;
    private Encoder rearLeftEncoder;
    private Encoder frontRightEncoder;
    private Encoder rearRightEncoder;

    private ADXRS450_Gyro gyro;

    private MecanumDrivePoseEstimator poseEstimator;

    private boolean slowMode;
    private boolean fieldOriented;

    @Logged(name = "Drivebase Type", importance = Importance.INFO)
    public static final String kDrivebaseType = "Mecanum";

    public MecanumDrivetrain(Pose2d initialPose, boolean slowModeAtStartup, boolean fieldOrientedAtStartup) {
        super();

        frontLeft = new VictorSP(MecanumDriveConstants.kFrontLeftPWM);
        rearLeft = new VictorSP(MecanumDriveConstants.kRearLeftPWM);
        frontRight = new VictorSP(MecanumDriveConstants.kFrontRightPWM);
        rearRight = new VictorSP(MecanumDriveConstants.kRearRightPWM);

        drive = new MecanumDrive(frontLeft, rearLeft, frontRight, rearRight);

        frontLeftEncoder = new Encoder(
            MecanumDriveConstants.kFrontLeftEncoderChannelA, 
            MecanumDriveConstants.kFrontLeftEncoderChannelB
        );
        frontLeftEncoder.setDistancePerPulse(MecanumDriveConstants.kDistancePerPulseMeters);

        rearLeftEncoder = new Encoder(
            MecanumDriveConstants.kRearLeftEncoderChannelA, 
            MecanumDriveConstants.kRearLeftEncoderChannelB
        );
        rearLeftEncoder.setDistancePerPulse(MecanumDriveConstants.kDistancePerPulseMeters);

        frontRightEncoder = new Encoder(
            MecanumDriveConstants.kFrontRightEncoderChannelA, 
            MecanumDriveConstants.kFrontRightEncoderChannelB
        );
        frontRightEncoder.setDistancePerPulse(MecanumDriveConstants.kDistancePerPulseMeters);

        rearRightEncoder = new Encoder(
            MecanumDriveConstants.kRearRightEncoderChannelA,
            MecanumDriveConstants.kRearRightEncoderChannelB
        );
        rearRightEncoder.setDistancePerPulse(MecanumDriveConstants.kDistancePerPulseMeters);

        gyro = new ADXRS450_Gyro();

        poseEstimator = new MecanumDrivePoseEstimator(
            MecanumDriveConstants.kKinematics, 
            Rotation2d.fromDegrees(getGyroAngle()), 
            new MecanumDriveWheelPositions(
                getFrontLeftEncoderMeters(), 
                getFrontRightEncoderMeters(), 
                getRearLeftEncoderMeters(), 
                getRearRightEncoderMeters()
            ), 
            initialPose
        );

        slowMode = slowModeAtStartup;
        fieldOriented = fieldOrientedAtStartup;
    }

    @Override
    public void periodic() {
        poseEstimator.update(
            Rotation2d.fromDegrees(getGyroAngle()), 
            new MecanumDriveWheelPositions(
                getFrontLeftEncoderMeters(), 
                getFrontRightEncoderMeters(), 
                getRearLeftEncoderMeters(), 
                getRearRightEncoderMeters()
            )
        );
    }

    @Override
    public void resetPose(Pose2d pose) {
        poseEstimator.resetPosition(
            Rotation2d.fromDegrees(getGyroAngle()), 
            new MecanumDriveWheelPositions(
                getFrontLeftEncoderMeters(), 
                getFrontRightEncoderMeters(), 
                getRearLeftEncoderMeters(), 
                getRearRightEncoderMeters()
            ), 
            pose
        );
    }

    @Override
    @Logged(name = "Pose", importance = Importance.INFO)
    public Pose2d getPose() {
        return poseEstimator.getEstimatedPosition();
    }

    @Override
    public boolean isSlowMode() {
        return slowMode;
    }

    @Override
    public boolean isFieldOriented() {
        return fieldOriented;
    }

    @Override
    public ChassisSpeeds getChassisSpeeds() {
        return MecanumDriveConstants.kKinematics.toChassisSpeeds(
            new MecanumDriveWheelSpeeds(
                getFrontLeftEncoderMetersPerSecond(), 
                getFrontRightEncoderMetersPerSecond(), 
                getRearLeftEncoderMetersPerSecond(), 
                getFrontRightEncoderMetersPerSecond()
            )
        );
    }

    @Override
    public Command toggleSlowMode() {
        return runOnce(() -> {
            slowMode = !slowMode;
        });
    }

    @Override
    public Command toggleFieldOriented() {
        return runOnce(() -> {
            fieldOriented = !fieldOriented;
        });
    }

    @Override
    public Command drive(CommandXboxController controller) {
        return run(() -> {
            double xSpeedDelivered = 
                controller.getLeftY()
                * (slowMode ? MecanumDriveConstants.kSlowModeMultiplier : 1)
                * (MecanumDriveConstants.kInvertForwardBackward ? -1 : 1)
                * MecanumDriveConstants.kMaxSpeedMetersPerSecond;

            double ySpeedDelivered = 
                controller.getLeftX()
                * (slowMode ? MecanumDriveConstants.kSlowModeMultiplier : 1)
                * (MecanumDriveConstants.kInvertLeftRight ? -1 : 1)
                * MecanumDriveConstants.kMaxSpeedMetersPerSecond;

            double rotDelivered =
                controller.getRightX()
                * (slowMode ? MecanumDriveConstants.kSlowModeMultiplier : 1)
                * (MecanumDriveConstants.kInvertRotation ? -1 : 1)
                * MecanumDriveConstants.kMaxAngularSpeedRadiansPerSecond;

            ChassisSpeeds speeds = fieldOriented ?
                ChassisSpeeds.fromFieldRelativeSpeeds(xSpeedDelivered, ySpeedDelivered, rotDelivered, Rotation2d.fromDegrees(getGyroAngle())) :
                new ChassisSpeeds(xSpeedDelivered, ySpeedDelivered, rotDelivered);

            drive.driveCartesian(
                speeds.vxMetersPerSecond / MecanumDriveConstants.kMaxSpeedMetersPerSecond, 
                speeds.vyMetersPerSecond / MecanumDriveConstants.kMaxSpeedMetersPerSecond, 
                speeds.omegaRadiansPerSecond / MecanumDriveConstants.kMaxAngularSpeedRadiansPerSecond
            );
        });
    }

    @Override
    public Command drive(ChassisSpeeds speeds) {
        return run(() -> {
            drive.driveCartesian(
                speeds.vxMetersPerSecond / MecanumDriveConstants.kMaxSpeedMetersPerSecond, 
                speeds.vyMetersPerSecond / MecanumDriveConstants.kMaxSpeedMetersPerSecond, 
                speeds.omegaRadiansPerSecond / MecanumDriveConstants.kMaxAngularSpeedRadiansPerSecond
            );
        });
    }

    @Override
    public Command drive(Pose2d targetPose) {
        return new PrintCommand("Mecanum Drivetrain Drive to Pose Not Implemented");
    }

    @Override
    public Command stop() {
        return run(() -> {
            drive.driveCartesian(0, 0, 0);
        });
    }

    @Logged(name = "Gyro Angle Degrees", importance = Importance.DEBUG)
    public double getGyroAngle() {
        return gyro.getAngle() * (MecanumDriveConstants.kInvertGyro ? -1 : 1);
    }

    @Logged(name = "Front Left Encoder Meters", importance = Importance.DEBUG)
    public double getFrontLeftEncoderMeters() {
        return frontLeftEncoder.getDistance() * (MecanumDriveConstants.kInvertEncoders ? -1 : 1);
    }

    @Logged(name = "Rear Left Encoder Meters", importance = Importance.DEBUG)
    public double getRearLeftEncoderMeters() {
        return rearLeftEncoder.getDistance() * (MecanumDriveConstants.kInvertEncoders ? -1 : 1);
    }

    @Logged(name = "Front Right Encoder Meters", importance = Importance.DEBUG)
    public double getFrontRightEncoderMeters() {
        return frontRightEncoder.getDistance() * (MecanumDriveConstants.kInvertEncoders ? -1 : 1);
    }

    @Logged(name = "Rear Right Encoder Meters", importance = Importance.DEBUG)
    public double getRearRightEncoderMeters() {
        return rearRightEncoder.getDistance() * (MecanumDriveConstants.kInvertEncoders ? -1 : 1);
    }

    @Logged(name = "Front Left Encoder Meters Per Second", importance = Importance.DEBUG)
    public double getFrontLeftEncoderMetersPerSecond() {
        return frontLeftEncoder.getRate() * (MecanumDriveConstants.kInvertEncoders ? -1 : 1);
    }

    @Logged(name = "Rear Left Encoder Meters Per Second", importance = Importance.DEBUG)
    public double getRearLeftEncoderMetersPerSecond() {
        return rearLeftEncoder.getRate() * (MecanumDriveConstants.kInvertEncoders ? -1: 1);
    }

    @Logged(name = "Front Right Encoder Meters Per Second", importance = Importance.DEBUG)
    public double getFrontRightEncoderMetersPerSecond() {
        return frontRightEncoder.getRate() * (MecanumDriveConstants.kInvertEncoders ? -1 : 1);
    }

    @Logged(name = "Rear Right Encoder Meters Per Second", importance = Importance.DEBUG)
    public double getRearRightEncoderMetersPerSecond() {
        return rearRightEncoder.getRate() * (MecanumDriveConstants.kInvertEncoders ? -1 : 1);
    }
}
